/* Wing channel names limited to 16 chars.  X32 channel names limited to 12 chars */
a_consoles = [
    {
        "name"         : "FOH",
        "type"         : "wing",
        "master"       : "yes",
        "ip"           : "192.168.1.155",
        "channel_map"  : [25, 26, 27, 28, 29, 30, 31],
        "placeholder"  : ["RF 1", "RF 2", "RF 3", "RF 4", "RF 5", "RF 6", "Wired Vox"] 
    },
    {
        "name"         : "Monitor",
        "type"         : "x32",
        "ip"           : "192.168.1.129",
        "channel_map"  : [23, 24, 25, 26, 27, 28, 29],
        "bus_map"      : [1, 2, 3, 4, 5, 6, 7],
        "bus_prefix"   : ["A- ", "B- ", "C- ", "D- ", "E- ", "F- ", "G- "]
    }
]

// TODO: What happens if placeholders aren't set in the a_consoles config array? Does it error out? If so, need sanity checking for such.


function f_buildUI(o_parent) {
    // Get master console config. We'll use this later as a source for itterating through channels
    o_master = f_getMasterConsole();
    //console.log (o_master);

    var o_table, o_row, o_cell, o_input, o_button;
    var console_id, channel_id;

    /*** BEGIN "TABLE" ***/
    o_table = document.createElement("DIV");
    o_table.setAttribute("class", "rTable");

    /*** BEGIN HEADER ROW OF TABLE ***/
    o_row = document.createElement("DIV");
    o_row.setAttribute("class", "rTableRow");

    // For each console configured...
    for (console_id=0; console_id<a_consoles.length; console_id++) {
        
        // Create spacer/label cell
        o_cell = document.createElement("DIV");
        o_cell.setAttribute("class", "rTableCell");
        //o_cell.appendChild(document.createTextNode(""));

        // Append cell to row
        o_row.appendChild(o_cell);

        // Create cell for inputs
        o_cell = document.createElement("DIV");
        o_cell.setAttribute("class", "rTableHead");
        o_cell.appendChild(document.createTextNode(a_consoles[console_id].name));

        // Append cell to row
        o_row.appendChild(o_cell);

        // If console has busses mapped, created columns for those as well.
        if (a_consoles[console_id].bus_map !== undefined) {
            
            // Create spacer/label cell
            o_cell = document.createElement("DIV");
            o_cell.setAttribute("class", "rTableCell");
            //o_cell.appendChild(document.createTextNode(""));

            // Append cell to row
            o_row.appendChild(o_cell);

            // Create cell for column labels
            o_cell = document.createElement("DIV");
            o_cell.setAttribute("class", "rTableHead");
            o_cell.appendChild(document.createTextNode(a_consoles[console_id].name+" (busses)"));

            // Append cell to row
            o_row.appendChild(o_cell);
        }
    }
    
    // Append row to table
    o_table.appendChild(o_row);

    /*** BEGIN ROWS FOR CONFIGURED CHANNEL/BUS MAPPINGS ***/
    // For each channel configured on master console, create a table row
    for (channel_id=0; channel_id<o_master.channel_map.length; channel_id++) {
        o_row = document.createElement("DIV");
        o_row.setAttribute("class", "rTableRow");

        // For each console configured...
        for (console_id=0; console_id<a_consoles.length; console_id++) {

            // Create spacer/label column
            o_cell = document.createElement("DIV");
            o_cell.setAttribute("class", "rTableCell labelCell");
            o_cell.appendChild(document.createTextNode("Ch " + a_consoles[console_id].channel_map[channel_id]));

            // Append cell to row
            o_row.appendChild(o_cell);

            // Create column for inputs
            o_cell = document.createElement("DIV");
            o_cell.setAttribute("class", "rTableCell");
            
            // Input object for channel/bus names
            o_input = document.createElement("input");
            o_input.setAttribute("type", "text");
            o_input.setAttribute("class", "name");
            o_input.setAttribute("id", "channel_name_" + console_id + "_" + channel_id);
            o_input.setAttribute("console_id", console_id);
            o_input.setAttribute("channel_id", channel_id);
            // o_input.setAttribute("placeholder", "ch " + a_consoles[console_id].channel_map[channel_id]);
            o_input.setAttribute("placeholder", o_master.placeholder[channel_id]);
            o_input.setAttribute("maxlength", f_maxNameLengthById(console_id));     

            // Attach an event listener to call f_onChange() function when user changes value (Will trigger on an ENTER or just the field losing focus)
            o_input.addEventListener(
                'change',
                function() { f_onChange(this); }
            );
            
            // Append input to column
            o_cell.appendChild(o_input);
            
            // Append column to row
            o_row.appendChild(o_cell);

            // If console has busses mapped, created columns for those as well.
            if (a_consoles[console_id].bus_map !== undefined) {

                // Create spacer/label column
                o_cell = document.createElement("DIV");
                o_cell.setAttribute("class", "rTableCell labelCell");
                o_cell.appendChild(document.createTextNode("Bus " + a_consoles[console_id].bus_map[channel_id]));

                // Append cell to row
                o_row.appendChild(o_cell);

                // Create column for inputs
                o_cell = document.createElement("DIV");
                o_cell.setAttribute("class", "rTableCell");

                // Input object
                o_input = document.createElement("input");
                o_input.setAttribute("type", "text");
                o_input.setAttribute("class", "name");
                o_input.setAttribute("disabled", "true"); 
                o_input.setAttribute("id", "bus_name_" + console_id + "_" + channel_id);
                o_input.setAttribute("console_id", console_id);
                o_input.setAttribute("bus_id", a_consoles[console_id].bus_map[channel_id]);
                o_input.setAttribute("placeholder", "bus " + a_consoles[console_id].bus_map[channel_id]);
                o_input.setAttribute("maxlength", f_maxNameLengthById(console_id));

                // Attach an event listener to call f_onChange() function when user changes value (Will trigger on an ENTER or just the field losing focus)
                o_input.addEventListener(
                    'change',
                    function() { f_onChange(this); }
                );

                // Append input to column
                o_cell.appendChild(o_input);

                // Append column to row
                o_row.appendChild(o_cell);
            }
        }
        
        // Append row to table
        o_table.appendChild(o_row);
    }

    /*** BEGIN ROW FOR COPY BUTTONS AND PREFIX ENABLE CHECKBOX ***/
    o_row = document.createElement("DIV");
    o_row.setAttribute("class", "rTableRow");
    for (console_id=0; console_id<a_consoles.length; console_id++) {

        // Create spacer/label column
        o_cell = document.createElement("DIV");
        o_cell.setAttribute("class", "rTableCell");
        //o_cell.appendChild(document.createTextNode(""));
        o_row.appendChild(o_cell);

        // Create column for inputs
        o_cell = document.createElement("DIV");
        o_cell.setAttribute("class", "rTableCell buttonCell");

        // Create "Copy From" button
        o_button = document.createElement("A");
        o_button.setAttribute("class", "button");
        o_button.setAttribute("href", "javascript:void(0)");
        o_button.appendChild(document.createTextNode("Copy From"));
        o_button.setAttribute("console_id", console_id);
        o_button.addEventListener(
            'click',
            function() { f_copyFrom(this.attributes.console_id.value); }
        );

        // Append button to cell
        o_cell.appendChild(o_button);

        // Append an empty text node to create a space between buttons
        o_cell.appendChild(document.createTextNode(" "));

        // Create "Clr All" button
        o_button = document.createElement("A");
        o_button.setAttribute("class", "button");
        o_button.setAttribute("href", "javascript:void(0)");
        o_button.appendChild(document.createTextNode("Clr All"));
        o_button.setAttribute("console_id", console_id);
        o_button.addEventListener(
            'click',
            function() { f_clrAll(this.attributes.console_id.value); }
        );

        // Append button to cell
        o_cell.appendChild(o_button);

        // Append cell to row
        o_row.appendChild(o_cell);

        /* */
        // If console has busses mapped, create a checkbox to enable/disble prefixes in bus names.
        if (a_consoles[console_id].bus_map !== undefined) {

            // Create spacer/label column
            o_cell = document.createElement("DIV");
            o_cell.setAttribute("class", "rTableCell");
            //o_cell.appendChild(document.createTextNode(""));
            o_row.appendChild(o_cell);

            // Create column for prefix checkbox
            o_cell = document.createElement("DIV");
            o_cell.setAttribute("class", "rTableCell buttonCell");
            
            // Input object
            o_input = document.createElement("input");
            o_input.setAttribute("type", "checkbox");
            o_input.setAttribute("id", "bus_prefix_enable_" + console_id);
            o_input.setAttribute("console_id", console_id);

            // Append input to cell
            o_cell.appendChild(o_input);

            // Append text label
            o_cell.appendChild(document.createTextNode(" Enable Prefixes"));

            // Append cell to row
            o_row.appendChild(o_cell);

        }
        /* */
    }
    
    // Append row to table
    o_table.appendChild(o_row);

    // Append table to parent
    o_parent.appendChild(o_table);
   
}

function f_copyFrom(src_console_id) {
    //console.log ("Copy from...");
    //console.log (o);
    //console.log ("foo: " + o.attributes.console_id.value);
    //src_console_id = o.attributes.console_id.value;

    var channel_name, bus_name;
    
    // Copy new names to each console configured
    for (dest_console_id=0; dest_console_id<a_consoles.length; dest_console_id++) {
        //console.log ("src_console_id  = " + src_console_id);
        //console.log ("dest_console_id = " + dest_console_id);

        // For each configured channel on the source console, copy values to all other consoles.
        for (channel_id=0; channel_id<a_consoles[src_console_id].channel_map.length; channel_id++) {

            if (src_console_id != dest_console_id) {
                // Get name from source input field and truncate to max length for destination console

                channel_name = document.getElementById("channel_name_" + src_console_id + "_" + channel_id).value.substring(0, f_maxNameLengthById(dest_console_id));
                
                // Write copied name to INPUT field
                f_writeInputField ( { "console_id" : dest_console_id,
                                      "channel_id" : channel_id,
                                      "value"      : channel_name } );
                
                // Send copied name to console
                f_sendNameAPI ( { "console_id"  : dest_console_id,
                                  "channel_id"  : channel_id,
                                  "name"        : channel_name } );
            }

            // Only copy channel names if src_console_id and dest_console_id are not the same
            if (a_consoles[dest_console_id].bus_map !== undefined) {
                bus_name = document.getElementById("channel_name_" + src_console_id + "_" + channel_id).value.substring(0, f_maxNameLengthById(dest_console_id));
                if (document.getElementById("bus_prefix_enable_" + dest_console_id).checked) {
                    bus_name = a_consoles[dest_console_id].bus_prefix[channel_id] + bus_name;
                    bus_name = bus_name.substring(0, f_maxNameLengthById(dest_console_id));
                }
                
                // Write copied name to INPUT field
                f_writeInputField ( { "console_id" : dest_console_id,
                                      "bus_id"     : channel_id,
                                      "value"      : bus_name } );
                
                // Send copied name to console
                f_sendNameAPI ( { "console_id"  : dest_console_id,
                                  "bus_id"      : channel_id,
                                  "name"        : bus_name } );
                
            }
        }
    }
}

function f_clrAll(console_id) {
    if (confirm ("Clear all channels on " + a_consoles[console_id].name  + "?")) {
        for (var channel_id=0; channel_id<a_consoles[console_id].channel_map.length; channel_id++) {

            // Write blank name to INPUT field
            f_writeInputField ( { "console_id" : console_id,
                                  "channel_id" : channel_id,
                                  "value"      : '' } );
            
            // Send blank name to console
            f_sendNameAPI ( { "console_id"  : console_id,
                              "channel_id"  : channel_id,
                              "name"        : '' } );
        }
    }
}

function f_onChange(o) {
    //   this.attributes.console_id.value, this.attributes.channel_id.value, this.value
    console.log ("TODO: trim whitespace and do sanity checking on new channel name.");
    
    console.log ("console_id = " + o.attributes.console_id.value);
    console.log ("channel_id = " + o.attributes.channel_id.value);
    //var console_id = o.attributes.console_id.value;
    //var channel_id = o.attributes.channel_id.value;
    //var new_name = document.getElementById("channel_name_" + console_id + "_" + channel_id).value;
    //f_sendNameAPI (channel_id, new_name, console_id);
    f_sendNameAPI ( { "console_id"  : o.attributes.console_id.value,
                      "channel_id"  : o.attributes.channel_id.value,
                      "name"        : document.getElementById("channel_name_" + o.attributes.console_id.value + "_" + o.attributes.channel_id.value).value } );
}





function f_sendNameAPI(o) {
    console.log (o);
    /*
      This function will be passed an obect with the following keys and data:
      console_id : Numerical identifier of console intended to receive the updated name (See below for how to convert to IP address)
      channel_id : If "name" is a channel name, this key will be defined and will contain the numerical id of the intended channel **
      bus_id     : If "name" is a bus name, this key will be defined and will contain the numerical id of the intended bus **
      name       : Value of new channel or bus name

      ** Note that only channel_id *OR* bus_id will be defined. This is the mechanism intended to determine whether we are update a channel or bus name.



     */

    // Make some more user-friendly copies of the variables in play
    var console_ip   = a_consoles[o.console_id].ip;
    var console_type = a_consoles[o.console_id].type;    // We need this since OSC for channel/bus names are different between the X32 and Wing.
    var name         = o.name;
    var channel_num, bus_num;

    // If name is a channel name...
    if (o.channel_id !== undefined) {
        channel_num = a_consoles[o.console_id].channel_map[o.channel_id];
        console.log ("Sending name '" + name + "' to channel " + channel_num + " on console " + a_consoles[o.console_id].name +
                     "(" + a_consoles[o.console_id].type + ") at " + console_ip + "...");

        /****  PATRICK  ****/
            
        // Here is where you will want to insert a fetch() to send the updated channel name.
        
        // JS variables you'll need:
        // console_ip
        // channel_num              (Wing OSC: /ch/<channel_num>/name )
        // name                     ( X32 OSC: /ch/<channel_num>/config/name <name>)

		if (console_type== "x32") {
			fetch("wSetX32IpChNumChName", {method: "PUT", body: console_ip + "," + channel_num + "," + name});
//			fetch("ServerPutOSC", {method: "PUT", body: "192.168.1.129,10023,/ch/01/config/name ,s titi"});
		} else {
			fetch("wSetWingIpChNumChName", {method: "PUT", body: console_ip + "," + channel_num + "," + name});
//			fetch("ServerPutOSC", {method: "PUT", body: '192.168.1.155,2223,/ ,s ch.3.name=toto,.4.name=titi'});
		}
        
        
    // If name is a bus name...
    } else if (o.bus_id !== undefined) {
        bus_num = a_consoles[o.console_id].bus_map[o.bus_id];
        console.log ("Sending name '" + name + "' to bus " + bus_num + " on console " + a_consoles[o.console_id].name +
                     "(" + a_consoles[o.console_id].type + ") at " + console_ip + "...");

        /****  PATRICK  ****/
            
        // Here is where you will want to insert a fetch() to send the updated bus name.
        
        // JS variables you'll need:
        // console_ip
        // bus_num                  (Wing OSC: /bus/<bus_num>/name )
        // name                     ( X32 OSC: /bus/<bus_num>/config/name <name>)
		if (console_type== "x32") {
			fetch("wSetX32IpBusNumBusName", {method: "PUT", body: console_ip + "," + bus_num + "," + name});
//			fetch("ServerPutOSC", {method: "PUT", body: '192.168.1.129,10023,/bus/01/config/name ,s toto'});
		} else {
			fetch("wSetWingIpBusNumBusName", {method: "PUT", body: console_ip + "," + bus_num + "," + name});
//			fetch("ServerPutOSC", {method: "PUT", body: '192.168.1.155,2223,/ ,s ch.3.name=toto,.4.name=titi'});
		}
    }
}


function f_getNames() {
    // This function will be called once when the page is initially loaded. I may opt to impliment
    // a "refresh" button at a later time, which will also call this function.
    // The purpose of this function is to get the current channel and bus names from each configured
    // console and then write them to the appopriate name fields using the f_writeInputField() function.

    // It would certainly be more efficient to do one call for each console
    
    
    // For each console configured...
    for (var console_id=0; console_id<a_consoles.length; console_id++) {

        // Get master console config. We'll use this later as a source for itterating through channels
        o_master = f_getMasterConsole();
        //console.log (o_master);

        var channel_num, bus_num;
        var name = '';
        var console_ip = a_consoles[console_id].ip;
        var console_type = a_consoles[console_id].type;    // We need this since OSC for channel/bus names are different between the X32 and Wing.

        // Get and display channel names for console
        for (var channel_id=0; channel_id<o_master.channel_map.length; channel_id++) {
            channel_num = a_consoles[console_id].channel_map[channel_id];
            console.log ("console_id = " + console_id);
            console.log ("console_ip = " + console_ip);
            console.log ("channel_id = " + channel_id);
            console.log ("actual channel number= " + channel_num);

            /****  PATRICK  ****/
            
            // Here is where you will want to insert a fetch() to get the current channel
            // name using the variables "console_ip" and "channel". The resulting name
            // should be stored to the "name" variable for display by the f_writeInputField() function.

            // JS variables you'll need:
            // console_ip
            // channel              ( X32 OSC: /ch/<channel_num>/config/name )
			//                      (Wing OSC: /ch/<channel_num>/name )
			//var ppp;
			if (console_type == "x32") {
				fetchChData('wGetX32IpChNameFromNum', console_ip + "," + channel_num, console_id, channel_id);
//				fetchChData('ServerPostOSC', '192.168.1.129,10023,/ch/02/config/name', console_id, channel_id);
			} else {
				fetchChData('wGetWingIpChNameFromNum', console_ip + "," + channel_num, console_id, channel_id);
//				fetchChData('ServerPostOSC', "192.168.1.155,2223,/ch/1/eq ,s *", console_id, channel_id);
			}  
        }
        
        // If console has busses configured, get and display bus names
        if (a_consoles[console_id].bus_map !== undefined) {
            console.log ("busses...");
            for (var bus_id=0; bus_id<o_master.channel_map.length; bus_id++) {
                bus_num = a_consoles[console_id].bus_map[bus_id];
                console.log ("console_id = " + console_id);
                console.log ("console_ip = " + console_ip);
                console.log ("bus_id = " + bus_id);
                console.log ("actual bus number = " + bus_num);

                /****  PATRICK  ****/
                
                // Here is where you will want to insert a fetch() to get the current bus
                // name using the variables "console_ip" and "bus". The resulting name
                // should be stored to the "name" variable for display by the f_writeInputField() function.

                // JS variables you'll need:
                // console_ip
                // bus_num                  ( X32 OSC: /bus/<bus_num>/config/name )
				//                      	(Wing OSC: /bus/<channel_num>/name )
				if (console_type == "x32") {
					fetchBusData('wGetX32IpBusNameFromNum', console_ip + "," + bus_num, console_id, bus_id);
//					fetchChData('ServerPostOSC', "192.168.1.129,10023,/bus/02/config/name", console_id, channel_id);
				} else {
					fetchBusData('wGetWingIpBusNameFromNum', console_ip + "," + bus_num, console_id, bus_id);
				}           
            }
        }
        
    }

}


async function fetchChData(url, data, co_id, ch_id) {
	const response = await fetch(url, {method: 'POST', body: data});

	if (!response.ok) {
		throw new Error(`HTTP error! status: ${response.status}`);
	}
	//console.log(responseText);
	f_writeInputField({ "console_id" : co_id,
                        "channel_id" : ch_id,
                        "value"      : await response.text()});
}


async function fetchBusData(url, data, co_id, bu_id) {
	const response = await fetch(url, {method: 'POST', body: data});

	if (!response.ok) {
		throw new Error(`HTTP error! status: ${response.status}`);
	}

	const responseText = await response.text();
	//console.log(responseText);
	f_writeInputField({ "console_id" : co_id,
                        "bus_id"     : bu_id,
                        "value"      : responseText});
}


function fUpdate() {
	 
f_writeInputField( { "console_id" : console_id,
                                 "channel_id" : channel_id,
                                 "value"      : name      
								 } );
}
		
		
function f_writeInputField(o) {
    //console.log (o);
    if (o.channel_id !== undefined) {
        document.getElementById('channel_name_' + o.console_id + '_' + o.channel_id).value = o.value;
    } else if (o.bus_id !== undefined) {
        document.getElementById('bus_name_' + o.console_id + '_' + o.bus_id).value = o.value;
    }
}


// This function simply fills in all the channel fields with dummy data
function f_testNames() {
    //console.log ("test names 2");
    var o_input={}
    a_names =['Brenda', 'Tanner', 'Olivia', 'Elijah', 'Abby', 'Beth', 'Sherry',
              'Jacque', 'Laura', 'Michelle', 'Dakota', 'Grace', 'Maria', 'Eric']
    for (var console_id=0; console_id<a_consoles.length; console_id++) { // For each console configured
        for (var channel_id=0; channel_id<o_master.channel_map.length; channel_id++) { // For each channel
            f_writeInputField( { "console_id" : console_id,
                                 "channel_id" : channel_id,
                                 "value"      : a_names.shift() } );
        }
    }
}

// Returns config object of first console in a_consoles array that is declared as the master.
// The only check is an element named "master" is defined.
function f_getMasterConsole() {
    for (var console_id=0; console_id<a_consoles.length; console_id++) {
        if (a_consoles[console_id].type !== undefined) {
            return a_consoles[console_id];
        }
    }
    // What to do?
    // If we've gotten here, none of the configured consoles have been declared as the master.
    // Which means the config is broken... Alert() console to let user know?
}

// Returns max number of chars based on type of console by console ID.
function f_maxNameLengthById(console_id) {
    if (a_consoles[console_id].type.toUpperCase() == 'WING') {
        return 16; // Wing consoles support up to 16 chars
    } else {
        return 12; // X32 consoles support up to 12 chars
    }
}

function f_flashError () {
  var l_dwellTime = 100; 
  l_originalBGColor = document.body.style.backgroundColor;
  document.body.style.backgroundColor = "#FF2222";
  //document.body.style.backgroundColor = "#FFFFFF";
  setTimeout('document.body.style.backgroundColor = l_originalBGColor;', (l_dwellTime));
}

//f_buildUI(document.getElementsByTagName('BODY')[0]);
f_buildUI(document.getElementById('tableContainer'));

//f_testNames();
f_getNames();

//console.log (a_consoles);
//console.log (f_getMasterConsoleId());

// X32
// /ch/[01..32]/config/name
// /ch/[01..32]/config/color enum {OFF,RD,GN,YE,BL,MG,CY,WH}
// /bus/[01..16]/config/name

// Wing
// /ch/[1..40]/name
// /ch/[1..40]/col [1..12]
// /bus/[1..16]/config/name
